# coding: utf-8

"""

    Main API Functionality
    ~~~~~~~~~~~~~~~~~~~~~~

"""
# Imports
import pycountry
import numpy as np
import pandas as pd

from pprint import pprint
from warnings import warn
from datetime import datetime

# Support Tools
from easymoney.support_tools import mint
from easymoney.support_tools import min_max
from easymoney.support_tools import closest_date
from easymoney.support_tools import year_extract
from easymoney.support_tools import min_max_dates
from easymoney.support_tools import closest_value
from easymoney.support_tools import fast_date_range
from easymoney.support_tools import date_format_check
from easymoney.support_tools import sort_range_reverse

# Pycountry Wrap
from easymoney.pycountry_wrap import PycountryWrap

# options_tools
from easymoney.options_tools import options_ranking
from easymoney.options_tools import year_date_overlap
from easymoney.options_tools import alpha2_by_cpi_years

# Easy Pandas
from easymoney.easy_pandas import pandas_null_drop
from easymoney.easy_pandas import pandas_pretty_print

# Online Data Sources
from easymoney.sources.ecb_interface import ecb_xml_exchange_data
from easymoney.sources.world_bank_interface import world_bank_pull



class EasyPeasy(object):
    """

    Tools for Monetary Information and Conversions.

    :param precision: number of places to round to when returning results. Defaults to 2.
    :type precision: ``int``
    :param fall_back: if True, fall back to closest possible date for which data is available. Defaults to True.
    :type fall_back: ``bool``
    :param fuzzy_match_threshold: a threshold for fuzzy matching confidence (requires the ``fuzzywuzzy`` package).
                                  The value must be an number between 0 and 100. The *suggested* minimum values is 85.
                                  This will only impact attempts to match on natural names, e.g., attempting to match
                                  'Canada' by passing 'Canadian'. If True, a threshold of 90 will be set. Defaults to False.

                                  .. warning::

                                        Fuzzy matching may yield inaccurate results.

                                        Whenever possible, use terminology *exactly* as it appears in ``options()``.

    :type fuzzy_threshold: ``int``, ``float`` or ``bool``
    :param data_path: alternative path to the database file(s). Defaults to None.
    :type data_path: ``str``
    """
    # Fix: `EasyPeasy()` does not handle currencies like 'EEK' properly.
    #       They may not been appearing in options() correctly.
    #       See: _user_currency_input() below.

    def __init__(self, precision=2, fall_back=True, fuzzy_threshold=False, data_path=None):
        """

        Initialize the ``EasyPeasy()`` class.

        """
        self._precision = precision
        self._fall_back = fall_back

        fuzzy_search_threshold = fuzzy_threshold
        recommended_fuzzy_threshold = 90
        min_suggested_fuzzy_threshold = 85
        # Check fuzzy_threshold
        if fuzzy_threshold is not False and not isinstance(fuzzy_threshold, (float, int)):
            raise ValueError("`fuzzy_threshold` must be either `False`, a `float` or an `int`.")
        elif fuzzy_threshold is True:
            fuzzy_search_threshold = recommended_fuzzy_threshold
        elif fuzzy_threshold is not False and isinstance(fuzzy_threshold, (float, int)) and fuzzy_threshold <= 0:
            raise ValueError("`fuzzy_threshold` must be greater than 0.")
        elif fuzzy_threshold is not False and isinstance(fuzzy_threshold, (float, int)) and fuzzy_threshold > 100:
            raise ValueError("`fuzzy_threshold` must be less than 100.")
        elif fuzzy_threshold is not False and isinstance(fuzzy_threshold, (float, int)) \
                and fuzzy_threshold < min_suggested_fuzzy_threshold:
            warn("\nLow `fuzzy_threshold` values, such as %s, have an elevated "
                 "likelihood of innaccurate results." % (str(fuzzy_threshold)))

        path_to_data = data_path if isinstance(data_path, str) else None
        self._pycountry_wrap = PycountryWrap(path_to_data, fuzzy_search_threshold)
        self._pycountries_alpha_2 = set([c.alpha_2 for c in list(pycountry.countries)])

        # CPI Dict
        self._cpi_dict = world_bank_pull(return_as='dict')
        self.alpha2_cpi_record = alpha2_by_cpi_years(regions=self._pycountries_alpha_2, cpi_dictionary=self._cpi_dict)

        # Exchange Dict
        self._exchange_dict, self._ecb_currency_codes, self._currency_date_record = ecb_xml_exchange_data(return_as='dict')

        # Min max for _currency_date_record
        self._currency_date_record_range = {k: fast_date_range(v, '%d/%m/%Y') for k, v in self._currency_date_record.items()}

        # Column Order for options
        self._table_col_order = ['RegionFull', 'Region', 'Alpha2', 'Alpha3', 'Currencies',
                                 'InflationDates', 'ExchangeDates', 'Overlap']


    def _params_check(self, amount="void", pretty_print="void"):
        """

        Check `amount` and `pretty_print` have been passed valid values.

        :param amount: numeric amount
        :type amount: ``float`` or ``int``
        :param pretty_print: must be a boolean
        :type pretty_print: ``bool``
        """
        # Check amount is float
        if amount != "void" and not isinstance(amount, (float, int)):
            raise ValueError("amount must be numeric (intiger or float).")

        # Check pretty_print is a bool
        if pretty_print != "void" and not isinstance(pretty_print, bool):
            raise ValueError("pretty_print must be either True or False.")


    def region_map(self, region, map_to='alpha_2'):
        """

        Map a 'region' to any one of: ISO Alpha 2, ISO Alpha 3, it's Name or Offical Name.

            Examples:
                - ``EasyPeasy().region_map(region='CA', map_to='alpha_2')``       :math:`=` 'CA'
                - ``EasyPeasy().region_map(region='Canada', map_to='alpha_3')``   :math:`=` 'CAN'

        :param region: a 'region' in the format of a ISO Alpha2, ISO Alpha3 or currency code, as well as natural name.
        :type region: ``str``
        :param map_to: 'alpha_2', 'alpha_3' 'name' or 'offical_name'. Defaults to 'alpha_2'.
        :type map_to: ``str``
        :return: the desired mapping from region to ISO Alpha2.
        :rtype: ``str`` or ``tuple``
        """
        # handle currency transitions here.
        return self._pycountry_wrap.map_region_to_type(region=region, extract_type=map_to)


    def _cpi_years(self, region, warn=True):
        """

        Get Years for which CPI information is available for a given region.

        :param region: ISO Alpha 2 Code.
        :type region: ``str``
        :param warn: warn if data could not be obtained.
        :type warn: ``bool``
        :return: list of years for which CPI information is available.
        :rtype: ``list``
        """
        cpi_years_list = self.alpha2_cpi_record.get(region, [])

        if len(cpi_years_list):
            return cpi_years_list
        elif warn:
            raise KeyError("Could not obtain inflation (CPI) information for '%s' from the\n" \
                           "International Monetary Fund database currently cached." % (region))
        else:
            return None


    def _cpi_match(self, region, year):
        """

        Match region to the best possible year.

        :param region: region of the form allowed by `EasyPeasy().region_map()`
        :type region: ``str``
        :param year: a year for which CPI information is desired.
                     Can also be one of: 'oldest' or 'latest'.
        :type year: ``int`` or ``str``
        :return: best matching of CPI information for the given region w.r.t. the year supplied.
        :rtype: ``float``, ``int`` or ``str``
        """
        # Initialize
        fall_back_year = None
        natural_region_name = None

        # replace year_b if it is 'oldest' or 'latest'
        available_years = list(map(int, self._cpi_years(region)))
        error_msg = "\nInflation (CPI) data for %s in '%s' could not be obtained from the\n" \
                    "International Monetary Fund database currently cached."
        warn_msg = error_msg + "\nFalling back to %s."

        if year == 'oldest':
            return min(available_years)
        elif year == 'latest':
            return max(available_years)
        elif int(float(year)) not in available_years:
            if self._fall_back:
                natural_region_name = self._pycountry_wrap.map_region_to_type(region, 'name')
                fall_back_year = closest_value(float(year), available_years)
                warn(warn_msg % (year, natural_region_name, str(fall_back_year)))
                return fall_back_year
            else:
                raise AttributeError(error_msg % (year, natural_region_name))
        else:
            return year


    def _cpi_region_year(self, region, year):
        """

        Get the Consumer Price Index (CPI) in a given region for a given year.

        :param region: region of the form allowed by `EasyPeasy().region_map()`
        :type region: ``str``
        :param year: a year for which CPI information is desired.
                     Can also be one of: 'oldest' or 'latest'.
        :type year: ``int`` or ``str``
        :return: CPI for a given year.
        :rtype: ``float``
        """
        cpi = self._cpi_dict.get(str(int(float(year))), {}).get(self.region_map(region, 'alpha_2'), None)
        if cpi is not None:
            return float(cpi)
        else:
            raise KeyError("Could not obtain inflation information for '%s' in '%s'." % (str(region), str(year)))


    def inflation(self, region, year_a, year_b=None, return_raw_cpi_dict=False, pretty_print=False):
        """

        Calculator to compute the inflation rate from Consumer Price Index (CPI) information.

        Inflation Formula:

        .. math:: Inflation_{region} = \dfrac{c_{1} - c_{2}}{c_{2}} \cdot 100

        | :math:`where`:
        |   :math:`c_{1}` = CPI of the region in *year_b*.
        |   :math:`c_{2}` = CPI of the region in *year_a*.

        :param region: a region.
        :type region: ``str``
        :param year_a: start year.
        :type year_a: ``int``
        :param year_b: end year. Defaults to None -- can only be left to this default if *return_raw_cpi_dict* is True.
        :type year_b: ``int``
        :param return_raw_cpi_dict: If True, returns the CPI information in a dict. Defaults to False.
        :type return_raw_cpi_dict: ``bool``
        :param pretty_print: if True, pretty prints the result otherwise returns the result as a float. Defaults to False.
        :type pretty_print: ``bool``
        :return: (a) the rate of inflation between year_a and year_h.

                 (b) a dictionary of CPI information with the years as keys, CPI as values.
        :rtype: ``float``, ``dict`` or ``NaN``
        """
        # Check Params
        self._params_check(pretty_print=pretty_print)

        # Map Region to its alpha 2 code
        mapped_region = self.region_map(region, 'alpha_2')

        # Set to_year
        to_year = self._cpi_match(mapped_region, year_b) if year_b is not None else None

        # Set from_year
        if year_a is not None:
            from_year = self._cpi_match(mapped_region, year_a)
        else:
            raise ValueError("year_a cannot be NoneType.")

        # Get the CPI for to_year and year_a
        c1 = self._cpi_region_year(mapped_region, to_year) if to_year is not None else None
        c2 = self._cpi_region_year(mapped_region, from_year)

        # Return dict, if requested
        if return_raw_cpi_dict != False:
            d = dict(zip(filter(None, [to_year, from_year]), filter(None, [c1, c2])))
            if return_raw_cpi_dict == 'complete':
                return d, {"year_a": from_year, "year_b": to_year}
            elif return_raw_cpi_dict == True:
                return d

        # Compute the rate of Inflation (and round).
        if (any([pd.isnull(c1), pd.isnull(c2)])) or (c2 == 0.0):
            return np.NaN
        else:
            rate = round(((c1 - c2) / float(c2)) * 100, self._precision)

        # Return or Pretty Print.
        if not pretty_print:
            return rate
        else:
            print(rate, "%")


    def inflation_calculator(self, amount, region, year_a, year_b, pretty_print=False):
        """

        Adjusts a given amount of money for inflation.

        :param amount: a monetary amount.
        :type amount: ``float`` or ``int``
        :param region: a geographical region.
        :type region: ``str``
        :param year_a: start year.
        :type year_a: ``int``
        :param year_b: end year.
        :type year_b: ``int``
        :param pretty_print: if True, pretty prints the result otherwise returns the result as a float.
                             Defaults to False.
        :type pretty_print: ``bool``
        :return: :math:`amount \cdot inflation \space rate`.
        :rtype: ``float`` or ``NaN``
        """
        # Check Params
        self._params_check(amount, pretty_print)

        # Input checking
        if year_a == year_b:
            return mint(amount, self._precision, self.region_map(region, map_to='currency_alpha_3'), pretty_print)

        # Get the CPI information
        inflation_dict, years = self.inflation(region, year_a, year_b, return_raw_cpi_dict='complete')

        # Block division by zero
        if inflation_dict[years['year_a']] == 0:
            warn("Problem obtaining required inflation information.")
            return np.NaN

        # Scale w.r.t. inflation.
        adjusted_amount = inflation_dict[years['year_b']] / float(inflation_dict[years['year_a']]) * amount

        # Print or Return
        return mint(adjusted_amount, self._precision, self.region_map(region, map_to='currency_alpha_3'), pretty_print)


    def _exchange_dates(self, currencies, min_max_rslt=False, date_format='%d/%m/%Y'):
        """

        Get all dates for which there is data for a given list of currencies

        :param currencies: a list of currencies
        :type currencies: ``list``
        :param min_max_rslt: compute the earliest and latest date for which exchange rate information is available.
                             Defaults to False.
        :type min_max_rslt: ``bool``
        :param date_format: a data format. Defaults to '%d/%m/%Y'.
        :type date_format: ``str``
        :return:
        :rtype: 1D ``list`` or 2D ``list``
        """
        d = self._currency_date_record_range if min_max_rslt else self._currency_date_record
        dates = [d.get(c.upper(), None) for c in currencies]

        # Remove None
        dates = list(filter(None, dates))
        if not len(dates):
            return None

        return dates[0] if len(dates) == 1 else dates


    def _user_currency_input(self, currency_or_region):
        """

        Converts User supplied reference to a currency into an actual ISO Alpha 3 Currency Code.

        :param currency_or_region: reference to a currency
        :type currency_or_region: ``str``
        :return: ISO Alpha 3 Currency Code
        :rtype: ``pycountry object``
        """
        # Note: 'temp. fix' has been added to handle currencies like 'EEK'.
        #        This capability should be integrated into region_map() in the future.
        try:
            return pycountry.currencies.lookup(currency_or_region).alpha_3
        except:
            if currency_or_region in self._ecb_currency_codes: # temp fix
                return currency_or_region
            else:
                return self.region_map(currency_or_region, "currency_alpha_3")


    def _base_cur_to_lcu(self, currency, date):
        """

        Convert from a base currency (Euros) to a local currency unit, e.g., CAD.

        :param currency: a currency code or region
        :type currency: ``str``
        :param date: date of allowed form (currently limited to DD/MM/YYYY).
        :type date: ``str``
        :return: exchange_rate w.r.t. the Euro as a base currency.
        """
        error_msg = "\nCould not obtain the exchange rate for '%s' on %s from the\n" \
                    "European Central Bank database currently cached."
        warn_msg = error_msg + "\nFalling back to %s."

        # Handle Base Currency
        if currency.upper() == 'EUR':
            return 1.0

        available_data = self._currency_date_record.get(currency, None)
        if available_data == None:
            raise AttributeError("Data could not obtained for '%s' from the\n" \
                                 "European Central Bank database currently cached." % (currency))

        if date == 'oldest':
            exchange_date = self._currency_date_record_range.get(currency)[0]
        elif date == 'latest':
            exchange_date = self._currency_date_record_range.get(currency)[1]
        elif isinstance(date, str) and date not in available_data:
            if self._fall_back:
                exchange_date = closest_date(date, available_data)
                warn(warn_msg % (currency, date, exchange_date))
            else:
                raise AttributeError(msg % (currency, exchange_date))
        elif isinstance(date, str) and date_format_check(date, from_format="%d/%m/%Y"):
            exchange_date = date
        else:
            raise ValueError("Invalid Date Supplied. Dates must be of the form DD/MM/YYYY.")

        exchange_rate = self._exchange_dict.get(exchange_date, {}).get(currency, None)
        if not isinstance(exchange_rate, type(None)):
            return exchange_rate
        else:
            raise AttributeError(error_msg % (currency, exchange_date))


    def currency_converter(self, amount, from_currency, to_currency, date="latest", pretty_print=False):
        """

        Function to perform currency conversion based on, **not** directly reported from, data obtained
        from the European Central Bank (ECB).
        Base Currency: EUR.

        **Formulae Used**

        Let :math:`LCU` be defined as:

            .. math:: LCU(\phi_{EUR}, CUR) = ExchangeRate_{\phi_{EUR} → CUR}

        | :math:`where`:
        |   :math:`CUR` is some local currency unit.
        |   :math:`\phi = 1`, as in the ECB data.
        |
        That is, less formally:

            .. math:: LCU(\phi_{EUR}, CUR) = \dfrac{x \space \space CUR}{1 \space EUR}

        :math:`Thus`:

            From :math:`CUR_{1}` to :math:`CUR_{2}`:

                .. math:: amount_{CUR_{2}} = \dfrac{1}{LCU(\phi_{EUR}, CUR_{1})} \cdot LCU(\phi_{EUR}, CUR_{2}) \cdot amount_{CUR_{1}}

        :param amount: an amount of money to be converted.
        :type amount: ``float`` or ``int``
        :param from_currency: the currency of the amount.
        :type from_currency: ``str``
        :param to_currency: the currency the amount is to be converted into.
        :type to_currency: ``str``
        :param date: date of data to perform the conversion with. Dates must be of the form: ``DD/MM/YYYY``.
        :type date: ``str``
        :param pretty_print: if True, pretty prints the table otherwise returns the table as a pandas DataFrame. Defaults to False.
        :type pretty_print: ``bool``
        :return: converted currency.
        :rtype: ``float``
        """
        # Check Params
        self._params_check(amount, pretty_print)

        # to/from_currency --> Currency Alpha 3 Code
        ca3 = [self._user_currency_input(c) if c not in self._ecb_currency_codes else c for c in (to_currency, from_currency)]
        to_currency_fn, from_currency_fn = ca3

        # Return amount unaltered if self-conversion was requested.
        if from_currency_fn == to_currency_fn:
            return mint(amount, self._precision, to_currency_fn, pretty_print)

        if any(x == None or x is None for x in [to_currency_fn, from_currency_fn]):
            raise ValueError("Could not convert '%s' to '%s'." % (from_currency, to_currency))

        # from_currency --> Base Currency --> to_currency
        conversion_to_invert = self._base_cur_to_lcu(from_currency_fn, date)
        if conversion_to_invert == 0.0:
            raise ZeroDivisionError("Cannot converted from '%s' on %s." % (from_currency, date))
        converted_amount = (conversion_to_invert ** -1) * self._base_cur_to_lcu(to_currency_fn, date) * float(amount)

        # Return results (or pretty print)
        return mint(converted_amount, self._precision, to_currency_fn, pretty_print)


    def normalize(self
                  , amount
                  , region
                  , from_year
                  , to_year="latest"
                  , base_currency="EUR"
                  , exchange_date="latest"
                  , pretty_print=False):
        """

        | Convert a Nominal Amount of money to a Real Amount in the same, or another, currency.
        |
        | This requires both inflation (for *currency*) and exchange rate information (*currency* to *base_currency*).
          See ``options(info = 'all', overlap_only = True)`` for an exhaustive listing of valid values to pass to this method.
        |
        | Currency Normalization occurs in two steps:
        |       1. Adjust the currency for inflation, e.g., 100 (2010 :math:`CUR_{1}`) → x (2015 :math:`CUR_{1}`).
        |       2. Convert the adjusted amount into the *base_currency*.

        :param amount: a numeric amount of money.
        :type amount: ``float`` or ``int``
        :param currency: a region or currency.
                         Legal options: Region Name, ISO Alpha2, Alpha3 or Currency Code (see ``options()``).
        :type currency: ``str``
        :param from_year: a year. For valid values see ``options()``.
        :type from_year: ``int``
        :param to_year: a year. For valid values see ``options()``.
                        Defaults to 'latest' (which will use the most recent data available).
        :type to_year: ``str`` or ``int``
        :param base_currency:  a region or currency. Legal: Region Name, ISO Alpha2, Alpha3 or Currency Code
                               (see ``options()``). Defaults to 'EUR'.
        :type base_currency: ``str``
        :param pretty_print: Pretty print the result if True; return amount if False. Defaults to False.
        :type pretty_print: ``bool``
        :return: amount adjusted for inflation and converted into the base currency.
        :rtype: ``float``
        """
        # Check Params
        self._params_check(amount, pretty_print)

        exchange_year = year_extract(exchange_date)
        if exchange_date not in ['oldest', 'latest'] and not str(exchange_year).isdigit():
            warn("\n`exchange_date` is likely formatted improperly.")

            # Check delta between `to_year` and `exchange_date`.
            if abs(float(to_year) - float(exchange_year)) > 1: # note, this doesn't currently handle oldest v. latest.
                warn("\nThe year for which the input amount is being adjusted\n"
                     "for inflation is %s, whereas the exchange rate year is %s." % (str(to_year), str(exchange_year)))

        # Adjust input for inflation
        real_amount = self.inflation_calculator(amount, region, year_a=from_year, year_b=to_year)

        # Compute Exchange
        normalize_amount = self.currency_converter(real_amount, region, base_currency, date=exchange_date)

        # Return results (or pretty print)
        return mint(normalize_amount, self._precision, self._user_currency_input(base_currency), pretty_print)


    def _options_info_error(self, rformat):
        """

        Error Messages for invalid information requests.

        :param rformat: one of: 'list' or 'table'.
        :type rformat: ``str``
        """
        options_error_msg = "Invalid Information Request.\n" \
                            "Please supply:\n" \
                            " - 'exchange' to `info` for a %s of currencies for which exchange rates are available.\n" \
                            " - 'inflation' to `info` for a %s of countries for which inflation information is available." \
                            % (rformat, rformat)

        if rformat == 'list':
            append = "\nTo see a complete summary, please set `rformat` equal to 'table' and set `info` equal to 'all'."
        elif rformat == 'table':
            append = "\n - 'all' to `info` for a complete summary of available data."

        raise ValueError(options_error_msg + append)


    def _options_table(self, info, table_overlap_only=False, range_table_dates=True):
        """

        Tool to Generate Information for ``options()`` in table form.

        :param info: 'exchange', 'inflation' or 'all' ('all' requires rformat is set to 'table').
        :type info: ``str``
        :param table_overlap_only:
        :param table_overlap_only: when info is set to 'all', keep only those rows for which exchange rate and inflation data overlap.
        :type table_overlap_only: ``bool``
        :param range_table_dates: if True, only report the minimum and maximum date for which data is available;
                              if False, all dates for which data is available will be reported. Defaults to True.
        :type range_table_dates: ``bool``
        :return: dataframe summarizing databases currently cached by ``EasyPeasy()``.
        :rtype: ``Pandas DataFrame``
        """
        # Note: does not currently handle currency transitions

        # Use CurrencyRelationshipsDB as Base
        d = [i for i in list(self._pycountry_wrap.alpha2_currency_dict.items()) if i[0] in self._pycountries_alpha_2]
        options_df = pd.DataFrame(d).rename(columns={0 : "Alpha2", 1 : "Currencies"})

        # Add Names
        options_df['Region'] = [self._pycountry_wrap.map_region_to_type(c, 'name') for c in list(options_df['Alpha2'])]
        options_df['RegionFull'] = [self._pycountry_wrap.map_region_to_type(c, 'official_name') for c in list(options_df['Alpha2'])]

        # Add Alpha3
        options_df['Alpha3'] = [self._pycountry_wrap.map_region_to_type(c, 'alpha_3') for c in list(options_df['Alpha2'])]

        # Map available Inflation Data onto this Base
        options_df['InflationDates'] = options_df['Alpha2'].map(
              lambda x: sort_range_reverse(self._cpi_years(x, warn=False), ('reverse' if not range_table_dates else 'range')), 'ignore'
        )

        # Map available Exchange Rate Data onto this Base
        options_df['ExchangeDates'] = options_df['Currencies'].map(
            lambda x: self._exchange_dates(x, min_max_rslt=range_table_dates), na_action='ignore'
        )

        # Add Overlap
        options_df['Overlap'] = options_df.apply(
            lambda x: year_date_overlap(x['InflationDates'], x['ExchangeDates'], date_format="%d/%m/%Y"), axis=1
        )

        # Fill NaNs
        options_df = options_df.fillna(np.NaN)

        # Weight Rows by Completeness; Sort by Data Overlap and Alpha2 Code
        options_df['TempRanking'] = options_df.apply(lambda x: options_ranking(x['InflationDates'], x['ExchangeDates']), axis=1)
        options_df = options_df.sort_values(['TempRanking', 'Region'], ascending=[False, True])

        if table_overlap_only and info == 'all':
            options_df = pandas_null_drop(options_df, subset=['InflationDates', 'ExchangeDates'])
        elif table_overlap_only and info != 'all':
            warn("`table_overlap_only` can only take effect if `info` is equal to 'all'.")

        # Subset
        col_order = self._table_col_order
        if info == 'exchange':
            col_order = [c for c in self._table_col_order if c != 'InflationDates']
        elif info == 'inflation':
            col_order = [c for c in self._table_col_order if c != 'ExchangeDates']
        elif info != 'all':
            self._options_info_error('table')

        return options_df.drop('TempRanking', axis=1)[col_order].reset_index(drop=True)


    def _options_lists(self, info):
        """

        Tool to Generate Information for ``options()`` in list form.

        :param info: 'exchange', 'inflation' or 'all' ('all' requires rformat is set to 'table').
        :type info: ``str``
        :return: list of requested information.
        :rtype: ``list``
        """
        if info.strip().lower() not in ['exchange', 'inflation']:
            self._options_info_error('list')

        d = self._exchange_dict if info.strip().lower() == 'exchange' else self._cpi_dict

        full = [list(v.keys()) for k, v in d.items()]
        return sorted(set([i for s in full for i in s]))


    def options(self, info='all', rformat='table', pretty_print=True, table_overlap_only=False, range_table_dates=True):
        """

        An easy interface to explore all of the terminology EasyPeasy understands
        as well the dates for which data is available.

        :param info: 'exchange', 'inflation' or 'all' ('all' requires rformat is set to 'table').
        :type info: ``str``
        :param rformat: 'table' for a table or 'list' for just the currency codes, alone. Defaults to 'table'.
        :type rformat: ``str``
        :param pretty_print: if True, prints the list or table. If False, returns the list or table (as a Pandas DataFrame).
                             Defaults to True.
        :type pretty_print: ``bool``
        :param table_overlap_only: when info is set to 'all', keep only those rows for which exchange rate and inflation data overlap.
        :type table_overlap_only: ``bool``
        :param range_table_dates: if True, only report the minimum and maximum date for which data is available;
                              if False, all dates for which data is available will be reported. Defaults to True.
        :type range_table_dates: ``bool``
        :return: information table or list
        :rtype: ``Pandas DataFrame`` or ``list``
        """
        pretty_df = None
        if rformat == 'list':
            request = self._options_lists(info)
            return request if not pretty_print else pprint(request, width=65, compact=True)
        elif rformat == 'table':
            request = self._options_table(info, table_overlap_only, range_table_dates)
            if pretty_print:
                pretty_df = request.drop('RegionFull', axis=1)
                pretty_df['Currencies'] = pretty_df['Currencies'].str.join("; ")
                pandas_pretty_print(pretty_df, col_align={'Region' : 'left'})
            else:
                return request
        else:
            raise ValueError("`rformat` must be one of:\n"
                             " - 'list', for a list of the requested information.\n"
                             " - 'table', for a table (dataframe) of the requested information.")



















# coding: utf-8

"""

    European Central Bank Data
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
# Modules
import re
import requests
import pandas as pd

from collections import defaultdict
from easymoney.support_tools import date_sort
from easymoney.support_tools import date_reformat


def _ecb_data_frame(exchange_rate_dict):
    """

    Convert `exchange_rate_dict` into a Pandas DataFrame

    :param exchange_rate_dict: data harvest by `ecb_xml_exchange_data()`.
    :type exchange_rate_dict: ``dict``
    :return: dataframe of ECB exchange data.
    :rytpe: ``Pandas DataFrame``
    """
    # Convert to pandas dataframe
    df = pd.DataFrame.from_dict(exchange_rate_dict, orient='index')

    # convert date index --> column
    df['Date'] = df.index
    df.index = range(df.shape[0])

    # Melt the dataframe
    df = pd.melt(df, id_vars=['Date'], value_vars=[d for d in df.columns if d != 'Date'])
    df.rename(columns={"variable": "Currency", "value": "Rate"}, inplace=True)

    # Drop NaNs
    df = df[pd.notnull(df['Rate'])]

    # Add Temp. Date Column
    df['DateTemp'] = pd.to_datetime(df['Date'], format='%d/%m/%Y')

    # Sort by currency code
    df.sort_values(['DateTemp', 'Currency'], ascending=[1, 1], inplace=True)

    # Drop DateTemp
    del df['DateTemp']

    # Refresh index and Return
    return df.reset_index(drop=True)


def ecb_xml_exchange_data(return_as='dict', ecb_extension="stats/eurofxref/eurofxref-hist.xml"):
    """

    | This tool harvests XML data European Central Bank via their generously provided API.
    | Expects the follwing in the XML data: 'time', 'currency' and 'rate'.
    | Returns either a Pandas DataFrame or nested dictionary of the form: ``{time: {currency: rate}}``.
    | Please do not write procedures that slam their servers.

    :param return_as: 'dict' for dictionary (nested); 'df' for Pandas DataFrame OR 'both' for both a dict and DataFrame.
    :type return_as: ``str``
    :param ecb_extension: URL to the exchange rate XML data on ``"http://www.ecb.europa.eu"``.
                          Defaults to ``'/stats/eurofxref/eurofxref-hist.xml'``.
    :type ecb_extension: ``str``
    :return: exchange rate with EUR as the base-currency.
    :rtype: ``dict`` or ``Pandas DataFrame``
    """
    # Currency Codes
    all_currency_codes = ['EUR']

    # Constuct the URL to the XML data on the ECB's website.
    xmlpath = "http://www.ecb.europa.eu/" + ecb_extension

    # Request the data from the sever.
    url_request = requests.get(xmlpath)

    # Convert the XML to str, split on 'Cube><Cube' and remove the first element.
    parsed_xml = str(url_request.content).split("Cube><Cube")[1:]

    # Initialize the exchange rate dict
    exchange_rate_dict = dict()

    # Initialize dict to track dates by currency
    currency_date_record = defaultdict(set)

    # Iterate though the parsed XML
    for j in parsed_xml:
        date = re.findall(r'time="(.*?)"', j)[0]
        ccodes = re.findall(r'currency="(.*?)" rate=', j)
        rates = re.findall(r'rate="(.*?)"', j)

        # Reformat Date
        reformatted_date = date_reformat(date, from_format="%Y-%m-%d")

        # Build Exchange Rate Dict
        exchange_rate_dict[reformatted_date] = dict(zip(ccodes, [float(x) for x in rates]))

        # Track Currency Codes
        all_currency_codes += [c for c in ccodes if c not in all_currency_codes]

        # Track Dates
        for c in ccodes:
            currency_date_record[c].add(reformatted_date)

    # Sort currency_date_record
    currency_date_record_sorted = {k: date_sort(v) for k, v in currency_date_record.items()}
    currency_date_record_sorted['EUR'] = date_sort(list(exchange_rate_dict.keys()))

    # return as dict
    if return_as == 'dict':
        return exchange_rate_dict, all_currency_codes, currency_date_record_sorted
    elif return_as == 'data_frame':
        return _ecb_data_frame(exchange_rate_dict), all_currency_codes
    elif return_as == 'both':
        return exchange_rate_dict, _ecb_data_frame(exchange_rate_dict), all_currency_codes, currency_date_record_sorted
    else:
        raise ValueError("`return_as` must be one of: 'dict', 'data_frame' or `both`.")

ecb_currency_to_alpha2_dict = {   "CYP": "CY"
                                , "EEK": "EE"
                                , "LTL": "LT"
                                , "LVL": "LV"
                                , "MTL": "MT"
                                , "ROL": "RO"
                                , "SIT": "SI"
                                , "SKK": "SK"
                                , "TRL": "TR"
}











# coding: utf-8

"""

    World Bank Group Data
    ~~~~~~~~~~~~~~~~~~~~~

"""
# Imports
import re
import wbdata
import numpy as np
import pandas as pd

from easymoney.easy_pandas import twoD_nested_dict


def _wb_rowwise_extractor(wb_row, dict_keys):
    """

    Extracts desired (see return) row information from the raw World Bank download.

    :param wb_row: a row of the raw World Bank Data (as a dataframe).
    :type wb_row: dict
    :return: dict with keys: Country, country id, Indicator, CPI, Year
    :rtype: dict
    """
    # dict with empty keys
    extracted_data_dict = dict.fromkeys(dict_keys)

    # Populate extracted_data_dict
    extracted_data_dict[dict_keys[0]]  =  wb_row['country']['value']
    extracted_data_dict[dict_keys[1]]  =  wb_row['country']['id']
    extracted_data_dict[dict_keys[2]]  =  wb_row['indicator']['value']
    extracted_data_dict[dict_keys[3]]  =  wb_row['value']
    extracted_data_dict[dict_keys[4]]  =  wb_row['date']

    return extracted_data_dict


def _wb_pull(dict_keys, raw_data):
    """

    Parse the information pulled from the World Bank's API.

    :param dict_keys:
    :return: DataFrame with the requested indicator information.
    :rtype: Pandas Dataframe
    """
    # Generate a list populated with dicts for each row of the raw data
    dictionaries = [_wb_rowwise_extractor(w, dict_keys) for w in raw_data]

    # Merge by key
    merged_dict = {k: [x[k] for x in dictionaries] for k in dictionaries[0]}

    # Merge into a single data frame
    data_frame = pd.DataFrame.from_dict(merged_dict)

    # Sort by region and year
    return data_frame.sort_values(['Alpha2', 'Year'], ascending = [1, 0]).reset_index(drop=True)


def world_bank_pull(value_true_name=None, indicator="FP.CPI.TOTL", return_as='data_frame'):
    """

    | Tool to harvest data for specific indicator from the World Bank Group via their generously provided API.
    | Extracts world bank information based on a specific indicator and returns a Pandas DataFrame.
    | Currently, this tools expects the following in the XML data:
                country, ISO alpha 2 code, an indicator, value name (to be replaced by value_true_name) and year.
    | Please do not write procedures that slam their servers.
    |
    | Acknowledgement: this tools is made possible by the `wbdata` package.¹
    |
    | ¹Sherouse, Oliver (2014). Wbdata. Arlington, VA.
    |

    :param value_true_name: reable name for the indicator. If None, this information will be extract from ``indicator``.
                            Defaults to None.
    :type value_true_name: ``str``
    :param indicator: World Bank Indicator. Defaults to "FP.CPI.TOTL".
    :type indicator: ``str``
    :param return_as: 'data_frame' or 'dict'
    :type return_as: ``str``
    :return: DataFrame with the requested indicator information or a dictionary
    :rtype: ``dict`` or ``Pandas DateFrame``
    """
    raw_data = wbdata.get_data(indicator)
    readable_name = value_true_name.split(".")[1] if value_true_name != None else value_true_name
    dict_keys = ['Country', 'Alpha2', 'Indicator', readable_name, 'Year']

    # Convert to DataFrame
    df = _wb_pull(dict_keys, raw_data)

    # Fill Empty Values with NaNs
    df = df.fillna(value=np.NaN).dropna(subset=['Year', 'Alpha2', readable_name]).reset_index(drop=True)

    if return_as == 'data_frame':
        return df
    elif return_as == 'dict':
        return twoD_nested_dict(df, 'Year', 'Alpha2', readable_name)
    elif return_as == 'both':
        return df, twoD_nested_dict(df, 'Year', 'Alpha2', readable_name)
    else:
        raise ValueError("Invalid option passed to `return_as`.")




































# coding: utf-8

"""

    Tools for Processing Included Databases
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""

# adds currency transitions
# Not complete. To finish this, pycountry_wrap and options() will need to be altered.

# Imports
import pandas as pd
import pkg_resources

from datetime import datetime
from collections import defaultdict
from easymoney.support_tools import cln


DEFAULT_DATA_PATH = pkg_resources.resource_filename('easymoney', 'sources/data')


def _path_selector(path_to_data):
    """

    Select path to the data.

    :param path_to_data: a path to the databases required by EasyMoney.
    :type path_to_data: ``str`` or ``None``
    :return: DEFAULT_DATA_PATH if `path_to_data` is None; else path_to_data.
    :rtype: ``str``
    """
    if isinstance(path_to_data, str):
        return (path_to_data if not path_to_data.endswith("/") else path_to_data[:-1])
    else:
        return DEFAULT_DATA_PATH


def currency_transitions(path_to_data, date_format="%d/%m/%Y"):
    """

    Tool to created a nested Dictionary of currency transitions of the form:
    ``{CountryAlpha2 : {'old': CUR_1, 'new': CUR_2, 'date': 'DD/MM/YYYY'}``.

    :param path_to_data: path to the 'CurrencyRelationshipsDB.csv' database.
    :type path_to_data: ``str``
    :return: a dictionary mapping alpha2 codes to currency codes.
    :rtype: ``dict``
    """
    # Read in the data
    transitions = pd.read_csv(_path_selector(path_to_data) + "/CurrencyTransitionDB.csv",
                              usecols=['Alpha2', 'OldCurrency', 'NewCurrency', 'Date'])

    transitions['Date'] = pd.Series(pd.to_datetime(transitions['Date'], format="%Y")).map(
        lambda x: x.strftime(date_format)
    )

    cols = zip(*[transitions[c] for c in ('Alpha2', 'OldCurrency', 'NewCurrency', 'Date')])
    return {country: {'old': old, 'new': new, 'date': date} for country, old, new, date in cols}


def _new_old_order(alpha2, currencies, transitions):
    """

    Sort regions with multiple known currencies into [NEW, OLD].
    Note: This will not handle multiple currency transitions.

    :param alpha2: an ISO alpha 2 region code.
    :type alpha2: ``str``
    :param currencies: an iterable of currencies.
    :type currencies: ``iterable``
    :param transitions: yeild of currency_transitions().
    :type transitions: ``dict``
    :return: list of the form [NEW CURRENCY, OLD CURRENCY].
    :rtype: ``list``
    """
    if len(currencies) > 1 and all(transitions[alpha2][i] in currencies for i in ['new', 'old']):
        new_old = [transitions[alpha2]['new'], transitions[alpha2]['old']]
        return new_old + [j for j in currencies if j not in new_old]
    else:
        return currencies


def currency_mapping_to_dict(path_to_data, date_format="%d/%m/%Y"):
    """

    Constructs a mapping between country ISO Alpha 2 codes and currency ISO Alpha 3 codes
    of the form:
    ``{CountryAlpha2 : {[Currencies]}``.

    :param path_to_data: path to the 'CurrencyRelationshipsDB.csv' database.
    :type path_to_data: ``str``
    :return: a dictionary mapping alpha2 codes to currency codes.
    :rtype: ``dict``
    """
    # Read in the data
    currency_mappings = pd.read_csv(_path_selector(path_to_data) + "/CurrencyRelationshipsDB.csv",
                                    usecols=['Alpha2', 'CurrencyCode'])

    # Initialize a default dict
    alpha2_currency = defaultdict(list)

    # Loop though to generate a
    for all_alpha2, currency in zip(*[currency_mappings[c] for c in ['Alpha2', 'CurrencyCode']]):
        if not any(pd.isnull(i) for i in [all_alpha2, currency]):
            for alpha2 in cln(all_alpha2).split(", "):
                if currency.upper() not in alpha2_currency[alpha2.upper()]:
                    alpha2_currency[alpha2.upper()].append(currency.upper())

    # Get currency transitions
    transitions = currency_transitions(path_to_data, date_format=date_format)

    # Add Currency Transitions
    for k, v in alpha2_currency.items():
        if k in transitions:
            for i in ['old', 'new']:
                if not transitions[k][i] in alpha2_currency[k]:
                    alpha2_currency[k].append(transitions[k][i])

    # Sort regions with currency transitions and Return
    return {k: _new_old_order(k, v, transitions) for k, v in alpha2_currency.items()}




























# coding: utf-8

"""

    Tools for EasyPeasy's options() Method
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
# Imports
import numpy as np
from datetime import datetime
from collections import defaultdict

from easymoney.support_tools import min_max
from easymoney.easy_pandas import items_null
from easymoney.support_tools import date_sort
from easymoney.support_tools import min_max_dates



def options_ranking(inflation, exchange):
    """

    Ranks ``options()`` table rows by their completeness.
    Rankings:
        3: inflation and exchange information
        2: exchange information
        1: inflation information
        0: neither inflation or exchange information

    :param inflation: date rage of data available for inflation data.
    :type inflation: ``list``
    :param exchange: date rage of data available for exchange data.
    :type exchange: ``list``
    :return: completeness of row
    :rtype: ``int``
    """
    if not items_null(inflation) and not items_null(exchange):
        return 3
    elif not items_null(exchange):
        return 2
    elif not items_null(inflation):
        return 1
    else:
        return 0


def year_date_overlap(years, full_dates, date_format="%d/%m/%Y"):
    """

    Get the overlap for a iterable of years and an iterable of full dates (as strings).

    :param range_a: years of the form (year_a, year_b).
    :type range_a: ``iterable``
    :param range_b: dates of the form (string_date_a, string_date_b).
    :type range_b: ``iterable``
    :return: date overlap.
    :rtype: ``list``
    """
    # Check inputs
    if any(items_null(i) for i in [years, full_dates]):
        return np.NaN

    # Convert input to datetimes
    year_datetimes = [datetime.strptime(dm + str(y), date_format) for y, dm in zip(min_max(years), ["01/01/", "31/12/"])]
    dates_datetimes = [datetime.strptime(d, date_format) for d in min_max_dates(full_dates, from_format=date_format)]

    # Get date floor
    date_floor = max([year_datetimes[0], dates_datetimes[0]])

    # Get date ceiling
    date_ceiling = min([year_datetimes[1], dates_datetimes[1]])

    # Return
    return [date_floor.strftime(date_format), date_ceiling.strftime(date_format)]


def alpha2_by_cpi_years(regions, cpi_dictionary):
    """

    Tool to construct a dictionary of the form:
    ``{CountryAlpha2: [Years for which CPI info is available]}``.

    :param regions: an iterable of ISO alpha 2 country codes.
    :type regions: ``iterable``
    :param cpi_dictionary: dictionary of CPI information as returned by ``world_bank_pull(return_as='dict')``.
    :param cpi_dictionary: ``dict``
    :return: dictionary of years for which CPI information is available for a given country.
    :rtype: ``dict``
    """
    # Note: not strictly for `EasyPeasy().options()` as this is also used in `EasyPeasy()._cpi_years()`.
    cpi_years_dict = defaultdict(list)
    for r in regions:
        for y in sorted(list(cpi_dictionary.keys()), reverse=True):
            if cpi_dictionary.get(y, None).get(r, None) != None:
                cpi_years_dict[r].append(y)

    return dict(cpi_years_dict)















# coding: utf-8

"""

    Tools for Processing Included Databases
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
# Imports
import pandas as pd
import pkg_resources

from collections import defaultdict
from easymoney.support_tools import cln


DEFAULT_DATA_PATH = pkg_resources.resource_filename('easymoney', 'sources/data')


def _path_selector(path_to_data):
    """

    Select path to the data.

    :param path_to_data: a path to the databases required by EasyMoney.
    :type path_to_data: ``str`` or ``None``
    :return: DEFAULT_DATA_PATH if `path_to_data` is None; else path_to_data.
    :rtype: ``str``
    """
    if isinstance(path_to_data, str):
        return (path_to_data if not path_to_data.endswith("/") else path_to_data[:-1])
    else:
        return DEFAULT_DATA_PATH


def currency_mapping_to_dict(path_to_data):
    """

    Constructs a mapping between country ISO Alpha 2 codes and currency ISO Alpha 3 codes.

    :param path_to_data: path to the 'CurrencyRelationshipsDB.csv' database.
    :type path_to_data: ``str``
    :return: a dictionary mapping alpha2 codes to currency codes
    :rtype: ``dict``
    """
    # Read in the data
    currency_mappings = pd.read_csv(_path_selector(path_to_data) + "/CurrencyRelationshipsDB.csv",
                                    usecols=['Alpha2', 'CurrencyCode'])

    # Initialize a default dict
    alpha2_currency = defaultdict(list)

    # Loop though to generate a
    for all_alpha2, currency in zip(*[currency_mappings[c] for c in ['Alpha2', 'CurrencyCode']]):
        if not any(pd.isnull(i) for i in [all_alpha2, currency]):
            for alpha2 in cln(all_alpha2).split(", "):
                if currency.upper() not in alpha2_currency[alpha2.upper()]:
                    alpha2_currency[alpha2.upper()].append(currency.upper())

    return alpha2_currency



























# coding: utf-8

"""

    Support Tools
    ~~~~~~~~~~~~~

"""
# Imports
import re
import pandas as pd
import dateutil.parser
from datetime import datetime


# ----------------------------------------------------------------------------------------------------------
# General
# ----------------------------------------------------------------------------------------------------------


def cln(i, extent = 1):
    """

    String white space 'cleaner'.

    :param i: input str
    :type i: ``str``
    :param extent: 1 --> all white space reduced to length 1; 2 --> removal of all white space.
    :return: cleaned string
    :rtype: ``str``
    """
    if isinstance(i, str) and i != "":
        if extent == 1:
            return re.sub(r"\s\s+", " ", i)
        elif extent == 2:
            return re.sub(r"\s+", "", i)
    else:
        return i


def min_max(iterable):
    """

    Get the min and max value of an iterable of numerics.

    :param iterable: any iterable data structure of numerics
    :type iterable: ``tuple``, ``list``, ``np.array``
    :return: (min(iterable), max(iterable))
    :rtype: ``tuple``
    """
    return (min(iterable), max(iterable))


def sort_range_reverse(iterable, action='range'):
    """

    Sort, reverse, or get range of an iterable.

    :param iterable: any iterable
    :type iterable: ``iterable``
    :param action: An action to perform on the iterable.
                   Must be one of: 'sort', 'reverse' or 'range'.
                   `iterable` must only contain numerics for 'range'.
                   Defaults to 'range'.
    :param action: ``str``
    :return: i. sorted or reversed iterable; ii. the min and max of the iterable.
    :rtype: ``list``
    """
    if not isinstance(iterable, list):
        return iterable
    elif action == 'sort':
        return sorted(iterable)
    elif action == 'reverse':
        return list(reversed(iterable))
    elif action == 'range':
        return list(min_max(iterable))
    else:
        return iterable


# ----------------------------------------------------------------------------------------------------------
# Mathematical
# ----------------------------------------------------------------------------------------------------------


def closest_value(value, list_of_values):
    """

    Get the closet value in a list of values.

    :param value: a numeric value.
    :type value: ``int``, ``float`` or ``datetime``
    :param list_of_values: an iterable.
    :type list_of_values: ``tuple``, ``list``, ``np.array``
    :return: item in `list_of_values` closet to `value`.
    :rtype: ``int``, ``float`` or ``datetime``
    """
    # Source: http://stackoverflow.com/a/12141207/4898004
    return min(list_of_values, key=lambda i: abs(i - value))


# ----------------------------------------------------------------------------------------------------------
# Dates
# ----------------------------------------------------------------------------------------------------------


def datetimer(date, from_format=None):
    """

    Converts a date string into a datetime.

    :param date: a date
    :type date: ``str``
    :param from_format: the format of `date`. If none, an attempt will be made to guess the format. Defaults to `None`.
    :type from_format: ``str`` or ``None``
    :return: a datetime object.
    :rtype: ``datetime``
    """
    formatter = datetime.strptime if from_format != None else dateutil.parser.parse
    return formatter(date, from_format)


def date_reformat(date, to_format='%d/%m/%Y', from_format=None):
    """

    Reformat a date to a given form.

    :param date: a date
    :type date: ``str``
    :param to_format: the new format of the date.
    :type to_format: ``str``
    :param from_format: the format of the string supplied.
    :type from_format: ``str``
    :return:
    """
    if not isinstance(date, str):
        return date
    return datetimer(date, from_format).strftime(to_format)


def date_sort(list_of_dates, from_format="%d/%m/%Y"):
    """

    Sort a list of datetimes.

    :param list_of_dates: iterable of date strings
    :type list_of_dates: ``iterable``
    :param from_format: the format of the strings in `list_of_dates`. Defaults to "%d/%m/%Y".
    :type from_format: ``str``
    :return: sorted date strings
    :rtype: ``list``
    """
    return sorted(list_of_dates, key=lambda x: datetime.strptime(x, from_format))


def min_max_dates(list_of_dates, from_format="%d/%m/%Y"):
    """

    Get the min and max date in a list of date strings.

    :param list_of_dates: list of date strings.
    :type list_of_dates: ``iterable``
    :param from_format: the format of the strings in `list_of_dates`. Defaults to "%d/%m/%Y".
    :type from_format: ``str``
    :return: (min(list_of_dates), max(list_of_dates))
    :rtype: tuple
    """
    pandas_datetime = pd.to_datetime(pd.Series(list_of_dates), format=from_format)
    return (pandas_datetime.min().strftime(from_format), pandas_datetime.max().strftime(from_format))


def fast_date_range(dates, from_format='%d/%m/%Y'):
    """

    Fast function to find the min and max date in an iterable of date strings.

    :param dates: iterable of strings
    :type dates: ``iterable``
    :param from_format: date format. Defaults to '%d/%m/%Y'.
    :type from_format: ``str``
    :return: [min(dates), max(dates)]
    :rtype: ``list``
    """
    return [i.strftime(from_format) for i in min_max([datetime.strptime(j, from_format) for j in dates])]


def closest_date(date, list_of_dates, from_format="%d/%m/%Y"):
    """

    Get the closet date (string) in a list of datestrings.

    :param date: a date.
    :type date: ``str``
    :param list_of_dates: an iterable of date strings.
    :type list_of_dates: ``iterable``
    :param from_format: the format of the strings in `list_of_dates`. Defaults to "%d/%m/%Y".
    :type from_format: ``str``
    :return: closest date in `list_of_dates` to `date`.
    :rtype: ``str``
    """
    list_of_datetimes = [datetimer(d, from_format) for d in list_of_dates]
    closest = closest_value(datetimer(date, from_format), list_of_datetimes)
    return closest.strftime(from_format)


def _canonical_datetime(date_format):
    """

    Converts a date format to a human-readable format.

    :param date_format: a python-recognized date format containing any of: '%d', '%m' or '%Y'.
    :type date_format: ``str``
    :return: %d --> DD;  %m --> MM;  %Y --> YYYY.
    :rtype: ``str``
    """
    for f, c in zip(['%d', '%m', '%Y'], ['DD', 'MM', 'YYYY']):
        date_format = date_format.replace(f, c)
    return date_format


def date_format_check(date, from_format="%d/%m/%Y"):
    """

    Check whether or not a date is of a given form.

    :param date_format: a date.
    :type date: ``str``
    :param from_format: a python-recognized date format.
    :type from_format: ``str``
    :return: True if `date` is of the form `from_format` else raises ValueError.
    :rtype: ``bool``
    """
    try:
        datetime.strptime(date, from_format)
        return True
    except ValueError:
        raise ValueError("Invalid date format.\n"
                         "Please supply a date of the form: %s." % (_canonical_datetime(from_format)))


def year_extract(date):
    """

    Extract year from a date.
    Assumes `date` to contain a year of the form: YYYY.

    :param date: a string containing a year
    :type date: ``str``
    :return: the first year found in string
    :rtype: ``str`` or ``None``
    """
    year = ""
    for c in str(date):
        if len(year) == 4:
            break
        if c.isdigit():
            year += c
        elif not c.isdigit() and len(year) < 4:
            year = ""
    return year if len(year) == 4 else None


# ----------------------------------------------------------------------------------------------------------
# Money Formatting
# ----------------------------------------------------------------------------------------------------------


def _money_formater(amount, precision, currency=''):
    """

    Formats money to be human-readable.

    :param amount: amount of money
    :type amount: ``float``, ``str`` or ``int``
    :param precision: number of place after the decimal to round to.
    :type amount: ``int``
    :param currency: A currency code to append to the result. Defaults to none ('').
    :type currency: ``str``
    :return: formated money
    :rtype: ``str``
    """
    return ('{:,.2f}'.format(round(float(amount), precision)) + ' ' + currency).strip()


def mint(amount, precision, currency='', pretty_print=False):
    """

    Print Money or Return it as a float.

    :param amount: amount of money
    :type amount: ``float``, ``str`` or ``int``
    :param precision: number of place after the decimal to round to.
    :param currency: A currency code to append to the result. Defaults to none ('').
    :type currency: ``str``
    :param pretty_print:
    :type pretty_print: ``bool``
    :return: money rounded to the requested precision or printed.
    :rtype: ``float`` or ``None``
    """
    if pretty_print:
        print(_money_formater(amount, precision, currency))
    else:
        return round(float(amount), precision)







#!/usr/bin/env python3

"""

    Database of which Nations use which Currency
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""

# Modules #
import os
import re
import copy
import numpy as np
import pandas as pd
import pkg_resources


from easymoney.easy_pandas import pandas_list_column_to_str


DEFAULT_DATA_PATH = pkg_resources.resource_filename('sources', 'data')
SINGLE_USE_PATH = pkg_resources.resource_filename('sources', 'single_use')


# Dict. for special cases in the self.currency_codes dataframe.
known_special_cases_dict = {
      "Democratic Republic of the Congo" : "CD"
    , "Cape Verde" : "CV"
    , "Egypt, auxiliary in Gaza Strip" : "EG"
    , "North Korea" : "KP"
    , "South Korea" : "KR"
    , "Laos" : "LA"
}


class CurrencyRelationships(object):
    """

    *SPECIFIC* set of tools for interacting with the included CurrencyCodesDB.csv and ISOAlphaCodesDB.csv databases.

    :param path_to_databases: an alterative path to the database files.
                              See ``easymoney.money.EasyPeasy(path_to_databases)`` for more.
    :type path_to_databases: str
    """


    def __init__(self, currency_codes_db = None, iso_alpha_codes_db = None):
        """
        Initialize the ``CurrencyRelationships()`` class.

        """
        # Initalize
        string_strip = lambda x: re.sub(r"\s\s+", " ", str(x).strip())

        # Import Currency Code Database.
        if currency_codes_db == None:
            self.currency_codes = pd.read_csv(SINGLE_USE_PATH + "/CurrencyCodesDB.csv", encoding = "ISO-8859-1")
        else:
            self.currency_codes = copy.deepcopy(currency_codes_db)

        # Clean the Locations column.
        self.currency_codes['Locations'] = self.currency_codes['Locations'].astype(str).map(string_strip)

        # Import Country Codes Database.
        if iso_alpha_codes_db == None:
            self.country_codes = pd.read_csv(DEFAULT_DATA_PATH + "/ISOAlphaCodesDB.csv", encoding = "ISO-8859-1", keep_default_na = False)
        else:
            self.country_codes = copy.deepcopy(iso_alpha_codes_db)

        # Dict relating alpha2 to alpha3.
        self.alpha2_to_alpha3_dict = dict(zip(self.country_codes['Alpha2'], self.country_codes['Alpha3']))

        # Dict relating a country's natural name (e.g., Canada) to its alpha2 code.
        self.name_to_alpha2_dict = dict(zip(self.country_codes['CountryName'], self.country_codes['Alpha2']))

    def alpha2_from_natural_name(self):
        """

        Parses lists that contain Natural Names (some of which *may* also have Alpha2 Codes in braces),
        e.g., Canada (CA)
              Cambodia
              China (CN)
                ...

        :return: see above description
        :rtype: list
        """
        # Note: this 'code soup' should be refactored in time.

        # Initialize
        region = None
        alpha2 = None
        alpha2_list = list()

        # Iterate through natural names
        for c in self.currency_codes['Locations'].tolist():
            if "(" not in c:
                region = [k for k in self.name_to_alpha2_dict.keys() if c.upper() in k.upper()]
                if c in known_special_cases_dict.keys():
                    alpha2_list.append(known_special_cases_dict[c])
                else:
                    if len(region) > 1:
                        region = [r for r in region if c.upper() == r.upper()]
                    if len(region) == 1:
                        alpha2 = [self.name_to_alpha2_dict[region[0]]]
                        if len(alpha2) == 1 and all(x not in [[x.strip() for x in alpha2]] for x in [np.nan, '']):
                            alpha2_list.append(alpha2)
                        else:
                            alpha2_list.append(np.NaN)
                    else:
                        alpha2_list.append(np.NaN)
            else:
                # Extract alpha2 codes from braces, e.g., 'Curacao (CW), Sint Maarten (SX)'.
                multi_codes = [x.replace(")", "").replace("(", "") for x in re.findall(r"\(\w+\)", c)]

                # Append
                alpha2_list.append(multi_codes[0]) if len(multi_codes) == 1 else alpha2_list.append(multi_codes)

        return alpha2_list

    def pandas_series_of_lists_cleanup(self, input_list):
        """

        Clean a list to allow it to be a Pandas series of lists.

        :param input_list: a list
        :type input_list: list
        :return: list of lists
        :rtype: list
        """
        # Clean up
        nan_replace_empty_list = [np.NaN if isinstance(c, list) and len(c) == 0 else c for c in input_list]

        # put strings into lists, if present and return.
        return [[c] if not isinstance(c, list) and isinstance(c, str) else c for c in nan_replace_empty_list]

    def relate(self):
        """

        Construct Pandas frame that relates Natural Name, Alpha2, Alpha3 and Currency Code.

        :returns: Natural Name, Alpha2, Alpha3 and Currency Code relationships.
        :rtype: Pandas DataFrame
        """
        # Get the alpha2 codes from the natural names
        alpha2_list = self.alpha2_from_natural_name()

        # Convert to a pandas acceptable format
        pandas_format_alpha2 = self.pandas_series_of_lists_cleanup(alpha2_list)

        # Use the self.currency_codes database as a framework
        relation_db = copy.deepcopy(self.currency_codes)

        # Correct for possible weird match and add as the Alpha2 column.
        relation_db['Alpha2'] = [['HK'] if i == ['HK', 'MO'] else i for i in pandas_format_alpha2]

        # Remove all bracketed information from the Locations column
        relation_db['Locations'] = relation_db['Locations'].map(lambda x: re.sub(r"\s\s+", "", re.sub(r"\(.*\)", "", x)).strip())

        # Remove rows with Alpha2 == nan
        relation_db = relation_db[pd.notnull(relation_db["Alpha2"])]
        
        # Refresh index
        relation_db.index = range(relation_db.shape[0])
        
        # Get Alpha3 code
        alpha3_list = list()
        for cc in relation_db.Alpha2.tolist():
            alpha3_list.append([self.alpha2_to_alpha3_dict[c] for c in cc])

        # Add Alpha3 Column
        relation_db['Alpha3'] = alpha3_list

        # Drop the Locations Column
        relation_db.drop('Locations', axis = 1, inplace = True)

        # Reorder
        relation_db = relation_db[['Currency', 'Code', 'Alpha2', 'Alpha3']]

        # Rename column
        relation_db.rename(columns = {'Code': 'CurrencyCode'}, inplace = True)

        return relation_db

# Save to Disk
# pandas_list_column_to_str(CurrencyRelationships().relate(), ['Alpha2', 'Alpha3']).to_csv('CurrencyRelationshipsDB.csv', index = False)














import os
from setuptools import setup, find_packages


def read(fname):
    try:
        return open(os.path.join(os.path.dirname(__file__), fname)).read()
    except:
        return 'Please see: https://github.com/TariqAHassan/EasyMoney.'


setup(
    name = "easymoney",
    version = "1.5.0",
    author = "Tariq A. Hassan",
    author_email = "laterallattice@gmail.com",
    description = ("Data Science Tools for Monetary Information and Conversions."),
    long_description = read('docs/README.rst'),
    license = "BSD",
    keywords = 'economics, finance, inflation, currency converter, data analysis, data science',
    url = "https://github.com/TariqAHassan/EasyMoney.git",
    download_url='https://github.com/TariqAHassan/EasyMoney/archive/v1.5.0.tar.gz',
    packages = find_packages(exclude = ["tests.*"
                                        , "tests"
                                        , "sources/single_use.*"
                                        , "sources/single_use"
                                        ]),
    package_data = {'easymoney': ['sources/data/*.csv'],},
    data_files = [('', ["LICENSE.txt"])],
    install_requires = ['numpy', 'pandas', 'pycountry', 'requests', 'wbdata'],
    classifiers = ["Development Status :: 5 - Production/Stable"
                   , "Natural Language :: English"
                   , "Intended Audience :: Science/Research"
                   , "Intended Audience :: Financial and Insurance Industry"
                   , "Programming Language :: Python :: 2.7"
                   , "Programming Language :: Python :: 3.3"
                   , "Programming Language :: Python :: 3.4"
                   , "Programming Language :: Python :: 3.5"
                   , "License :: OSI Approved :: BSD License"
    ],
    include_package_data = True
)





# coding: utf-8

"""

    EasyMoney Wrapper for Pycountry
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
# Import
import pycountry

from warnings import warn
from easymoney.sources.databases import currency_mapping_to_dict



class PycountryWrap(object):
    """

    Tools wrapping the `pycountry` module.

    :param path_to_data: path to the database file(s). Defaults to None.
    :type path_to_data: ``str``
    :param fuzzy_match_threshold: a threshold for fuzzy matching confidence (requires the ``fuzzywuzzy`` package).
                                  For more, see ``EasyPeasy()`` in the `money` module.
    :type fuzzy_threshold: ``int`` or ``float``
    """


    def __init__(self, path_to_data=None, fuzzy_threshold=False):
        """

        Initialize the ``PycountryWrap()`` class.

        """
        # Compute the dict mapping alpha2 codes to currencies
        self.alpha2_currency_dict = currency_mapping_to_dict(path_to_data)

        # Get a list of country names
        self.countries = [c.name for c in list(pycountry.countries)]

        # FuzzyWuzzy Settings
        self.fuzzy_threshold = fuzzy_threshold

        if fuzzy_threshold != False:
            try:
                from fuzzywuzzy import process
                self.extractOne = process.extractOne
            except:
                raise ImportError("\nTo use `fuzzy_threshold` please install `fuzzywuzzy` as well as\n"
                                  "`python-Levenshtein` for optimal performance.\n\n"
                                  " - python 2: $ pip install fuzzywuzzy python-Levenshtein\n"
                                  " - python 3: $ pip3 install fuzzywuzzy python-Levenshtein\n\n"
                                  "You may need to create a new Python instance for these changes to take effect.")


    def _country_extract(self, country, extract_type='alpha_2'):
        """

        Extract information from a given country in the `pycountry` database.

        :param country: a pycountry object to harvest `extract_type` from.
                        Recognized types of information: 'alpha_2', 'alpha_3', 'numeric', 'name' or 'official_name'.

        :type country: ``pycountry object``
        :param extract_type: The type of information to extract from `country`. Defaults to 'alpha_2'.
        :type extract_type: ``str``
        :return: a country's ISO alpha 2 code, ISO alpha 3 code, numeric code, name or official name.
        :rtype: ``str`` or ``None``
        """
        if extract_type == 'alpha_2':
            try:
                return country.alpha_2
            except:
                return None
        elif extract_type == 'alpha_3':
            try:
                return country.alpha_3
            except:
                return None
        elif extract_type == 'numeric':
            try:
                return country.numeric
            except:
                return None
        elif extract_type == 'name':
            try:
                return country.name
            except:
                return None
        elif extract_type == 'official_name':
            try:
                return country.official_name
            except:
                try:
                    return country.name
                except:
                    return None
        else:
            raise ValueError("invalid extract_type supplied")


    def _currency_extract(self, currency_name, extract_type='currency_alpha_3'):
        """

        Extract information from a given currency in the `pycountry` database.

        :param currency_name: a country.
        :type currency_name: ``str``
        :param extract_type: Defaults to 'currency_alpha_3'.
        :type extract_type: ``str``
        :return: an ISO Alpha 3 currency code.
        :rtype: ``pycountry object`` or ``None``
        """
        if extract_type == 'currency_alpha_3':
            try:
                return pycountry.currencies.get(alpha_3=currency_name).alpha_3
            except:
                return None
        elif extract_type == 'currency_numeric':
            try:
                return pycountry.currencies.get(alpha_3=currency_name).numeric
            except:
                return None

        elif extract_type == 'currency_name':
            try:
                return pycountry.currencies.get(alpha_3=currency_name).name
            except:
                return None
        else:
            raise ValueError("invalid extract_type supplied")


    def _fuzzy_search(self, term, options):
        """
        
        Fuzzy Searching tool via the `fuzzywuzzy` module.

        :param term: a search term.
        :type term: ``str``
        :param options: an iterable of options to match on.
        :type options: ``iterable``
        :return: best match match quality is less than the threshold.
        :rtype: ``str``
        """
        rslt = self.extractOne(term, options)
        return rslt[0] if rslt[1] >= self.fuzzy_threshold else None


    def _region_lookup(self, region):
        """
        
        Tool to find a region's `pycountry` object.

        :param region: a region.
        :type region: ``str``
        :return: a `pycountry` object for `region`.
        :rtype: ``pycountry object`` or ``None``
        """
        fuzzy_match = None
        try:
            return pycountry.countries.lookup(region)
        except:
            if self.fuzzy_threshold != False:
                fuzzy_match = self._fuzzy_search(region, self.countries)
                if fuzzy_match is not None:
                    return pycountry.countries.lookup(fuzzy_match)
                else:
                    return None
            else:
                return None


    def map_region_to_type(self, region, extract_type='alpha_2'):
        """

        Maps the input region to:
            region: 'alpha_2', 'alpha_3', 'name' or 'official_name'.
            currency: 'currency_alpha_3', 'currency_numeric' or 'currency_name'.

        :param region: any region
        :type region: ``str``
        :param extract_type: the type of information about `region` to extract. Must be one of the types listed above.
        :type extract_type: ``str``
        :return: `extract_type` information.
        :rtype: ``str`` or ``None``
        """
        rslt = self._region_lookup(region)
        if rslt is None:
            return None

        try:
            if 'currency_' not in extract_type.lower():
                return self._country_extract(rslt, extract_type.lower())

            else:
                # Get the alpha_3 country code
                alpha_2 = self._country_extract(rslt, 'alpha_2')

                # Look up
                currencies = self.alpha2_currency_dict[alpha_2]

                # Extract
                if len(currencies):
                    if len(currencies) > 1:
                        warn("Multiple currencies are used in %s, including: %s.\n"
                             "%s has been selected by default." % (region, ", ".join(currencies), currencies[0]))
                    return self._currency_extract(currencies[0], extract_type)
                else:
                    return None
        except:
            return None





















# coding: utf-8

"""

    Supporting Pandas Tools
    ~~~~~~~~~~~~~~~~~~~~~~~

"""

# Modules #
import copy
import warnings
import datetime
import numpy as np
import pandas as pd


def pstr(s):
    """

    Convert to any obect to a string using pandas.

    :param s: item to be converted to a string.
    :type s: ``any``
    :return: a string
    :rtype: ``str``
    """
    return pd.Series([s]).astype('unicode')[0]


def strlist_to_list(to_parse, convert_to_str_first=False):
    """

    Work around for using ``eval()`` for the following conversion: ``str(list)`` → ``list``.

    For example: ``"[1992, '221-21', 2102, 'apples']"`` → ``['1992', '221-21', '2102', 'apples']``.

    :param to_parse: a string of a list.
    :type to_parse: ``str``
    :param convert_to_str_first: convert to a string first (as a precaution). Defaults to False.
    :type convert_to_str_first: ``bool``
    :return: string of a list to an actual list.
    :return: ``list``
    """
    str_list = pstr(to_parse) if convert_to_str_first else to_parse
    return [i.strip().replace("'", "") for i in [j.split(",") for j in [str_list.replace("[", "").replace("]", "")]][0]]


def _pandas_dictkey_to_key_unpack(pandas_series, unpack_dict, convert_values_to_str = False):
    """

    Used to unpack ISO Alpha2 --> Alpha2 ISO code Pandas Series.

    :param pandas_series: a Series to replace the alpha2 codes with another set.
    :type pandas_series: ``Pandas Series``
    :param convert_values_to_str: convert the values to string (precaution).
    :type convert_values_to_str: ``bool``
    :param unpack_dict: a dict with the values coerced into strings.
    :type unpack_dict: ``dict``
    :return: a pandas series with the alpha2 values replaced with unpack_dict.values().
             ``NaN`` is used if a match cannot be found.
    :rtype: ``Pandas Series``
    """
    if convert_values_to_str:
        unpack_dict = {k: pstr(v) for k, v in unpack_dict.items()}

    return pandas_series.replace(unpack_dict).map(lambda x: np.NaN if 'nan' in pstr(x) else strlist_to_list(pstr(x)))


def _standard_pd_nester(data_frame, nest_col_a, nest_col_b, nest_col_c, keys_to_str = True):
    """

    Method to produce a nested dict from a large pandas dataframe.
    Reliable technique (although slow with large DataFrames).

    :param data_frame: see ``twoD_nested_dict()``.
    :type data_frame: ``Pandas DataFrame``
    :param nest_col_a: see ``twoD_nested_dict()``.
    :type nest_col_a: ``str``
    :param nest_col_b: see ``twoD_nested_dict()``.
    :type nest_col_b: ``str``
    :param nest_col_c: see ``twoD_nested_dict()``.
    :type nest_col_c: ``str``
    :param keys_to_str: see ``twoD_nested_dict()``.
    :type keys_to_str: ``bool``
    :return: nested dict of the form: ``{nest_col_a: {nest_col_b: nest_col_c}}``.
    :rtype: dict
    """
    # Initialize
    df_slice = None

    # Make columns that are to become keys strings.
    if keys_to_str:
        data_frame[nest_col_a] = data_frame[nest_col_a].astype(str).str.upper()
        data_frame[nest_col_b] = data_frame[nest_col_b].astype(str)

    # Create a dict from keys()
    nested_dict = dict.fromkeys(data_frame[nest_col_a].unique())
    for k in nested_dict.keys():
        df_slice = data_frame[data_frame[nest_col_a] == k]
        nested_dict[k] = dict(zip(df_slice[nest_col_b], df_slice[nest_col_c]))

    return nested_dict


def _fast_pd_nester(data_frame, nest_col_a, nest_col_b, nest_col_c, keys_to_str = True):
    """

    | This is a fast way to produce a nested dict from a large Pandas DataFrames.
    | Can handle DataFrames with several nest_col_a entries that are the same,
    | e.g.,
    |         nest_col_a    nest_col_b  nest_col_c
    | 0       1999-01-04      AUD          0
    | 1       1999-01-04      CAD          1
    | 2       1999-01-05      CHF          2
    | 3       1999-01-05      CYP          3
    |
    | WARNING: Not well-tested. This procedure may produce inaccuracies.

    :param data_frame: see ``twoD_nested_dict()``.
    :type data_frame: Pandas DataFrame
    :param nest_col_a: see ``twoD_nested_dict()``.
    :type nest_col_a: str
    :param nest_col_b: see ``twoD_nested_dict()``.
    :type nest_col_b: str
    :param nest_col_c: see ``twoD_nested_dict()``.
    :type nest_col_c: str
    :param keys_to_str: see ``twoD_nested_dict()``.
    :type keys_to_str: ``bool``
    :return: nested dict of the form {nest_col_a: {nest_col_b: nest_col_c}.
    :rtype: dict
    """
    # Group nest_col_b and nest_col_c by nest_col_a
    col_b_groupby = data_frame.groupby(nest_col_a)[nest_col_b].apply(lambda x: x.tolist()).reset_index()
    col_c_groupby = data_frame.groupby(nest_col_a)[nest_col_c].apply(lambda x: x.tolist()).reset_index()

    # Merge on nest_col_a
    grouped_data_frame = pd.merge(col_b_groupby, col_c_groupby, on = nest_col_a, how = 'outer')

    # Convert columns that will become keys to a string
    if keys_to_str:
        grouped_data_frame[nest_col_a] = grouped_data_frame[nest_col_a].astype(str)
        grouped_data_frame[nest_col_b] = grouped_data_frame[nest_col_b].map(lambda x: [pstr(i) for i in x])

    # Generate a dictionary by zipping the nest_col_b and nest_col_c columns
    grouped_data_frame['dict_zipped'] = grouped_data_frame[[nest_col_b, nest_col_c]].apply(
                                                                    lambda x: dict(zip(x[0], x[1])), axis = 1)

    # Return as a nested dict
    return dict(zip(grouped_data_frame[nest_col_a], grouped_data_frame['dict_zipped']))


def twoD_nested_dict(data_frame
                     , nest_col_a = None
                     , nest_col_b = None
                     , nest_col_c = None
                     , to_float = None
                     , to_int = None
                     , keys_to_str = True
                     , engine = 'standard'):
    """

    Generate a nested dictionary from the columns of a pandas dataframe.
    Defaults to using the first 3 columns.

    :param data_frame: a pandas dataframe.
    :type data_frame: ``Pandas DateFrame``
    :param nest_col_a: reference to column in the dataframe; to become the master key in dict.
                       Defaults to None (i.e., col 1).
    :type nest_col_a: ``str``
    :param nest_col_b: reference to column in the dataframe; to become the sub-key.
                       Defaults to None (i.e., col 2).
    :type nest_col_b: ``str``
    :param nest_col_c: reference to column in the dataframe; to become the value to corresponding to the sub-key.
                       Defaults to None (i.e., col 3).
    :type nest_col_c: ``str``
    :param to_float: a list items to float. Defaults to None.
    :type to_float: ``str``
    :param to_int: a list of the lists to convert to ints. Defaults to None.
    :type to_int: ``str``
    :param keys_to_str: Convert the columns that will become keys to strings. Default to True.
                        WARNING: will OVERRIDE *to_float* and *to_int* if they reference nest_col_a or nest_col_b.
    :type keys_to_str: ``bool``
    :param engine: 'standard' for a slower (but well-tested) method of generating a nested dict; 'fast' to employ a
                    speedy (but *not well-tested*) method for generating a nested dict. Default to 'standard'.
    :type engine: ``str``
    :return: nested dict of the form: ``{nest_col_a: {nest_col_b: nest_col_c}``.
    :rtype: ``dict``
    """
    if all(v is None for v in [nest_col_a, nest_col_b, nest_col_c]):
        nest_col_a = data_frame.columns[0]
        nest_col_b = data_frame.columns[1]
        nest_col_c = data_frame.columns[2]

    # Convert selected columns to float
    if to_float != None:
        for i in to_float: data_frame[i] = data_frame[i].astype(float)

    # Convert selected columns to int
    if to_int != None:
        for j in to_int: data_frame[j] = data_frame[j].astype(int)

    # Use one of two engines to generate the nested dict
    if engine == 'standard':
        return _standard_pd_nester(data_frame, nest_col_a, nest_col_b, nest_col_c, keys_to_str)

    elif engine == 'fast':
        return _fast_pd_nester(data_frame, nest_col_a, nest_col_b, nest_col_c, keys_to_str)


def pandas_list_column_to_str(data_frame, columns, join_on = ", ", bracket_wrap = False):
    """

    Tool for converting the columns in a Pandas DataFrame
    from pd.Series of lists into comma-seperated strings.

    :param data_frame: a dataframe.
    :type data_frame: ``Pandas DataFrame``
    :param columns: a list of columns in the DataFrame
    :type columns: ``list``
    :param join_on: a string to join on. Defaults to ", ".
    :type join_on: ``str``
    :return: a dataframe with the columns altered in the manner described above.
    :rtype: ``Pandas DataFrame``
    """
    df = copy.deepcopy(data_frame)
    for col in columns:
        df[col] = df[col].map(lambda x: join_on.join(x) if pstr(x) != 'nan' else x)
        if bracket_wrap:
            df[col] = df[col].map(lambda x: "[" + pstr(x) + "]" if pstr(x) != 'nan' else x)

    return df


def pandas_str_column_to_list(data_frame, columns):
    """

    Tool for converting the columns in a Pandas DataFrame
    from comma-seperated strings into a pd.Series of lists.

    :param data_frame: a dataframe.
    :type data_frame: ``Pandas DataFrame``
    :param columns: a list of columns in the dataframe.
    :type columns: ``list``
    :return: a dataframe with the columns altered in the manner described above.
    :rtype: ``Pandas DataFrame``
    """
    data_frame = copy.deepcopy(data_frame)
    for col in columns:
        data_frame[col] = data_frame[col].astype(str).map(lambda x: [i.strip() for i in x.split(",")])
    return data_frame


def type_in_series(series):
    """

    Return the types of objects in a Pandas Series.

    :param series: a series.
    :type series: ``Pandas Series``
    :return: list of the types in a series.
    :rtype: ``list``
    """
    return list(set([type(i).__name__ if pstr(i).strip() not in ['nan', ''] else 'nan' for i in series]))


def prettify_all_pandas_list_cols(data_frame, join_on = ", ", allow_nan = True, exclude = [], bracket_wrap = False):
    """

    Converts all columns with only lists to list-seperated-strings.

    :param data_frame: a dataframe.
    :type data_frame: ``Pandas DataFrame``
    :param nan: allow nans
    :type nan: ``bool``
    :param join_on: a string to join on. Defaults to ", ".
    :type join_on: ``str``
    :param exclude: columns to exclude. Defaults to an empty list (``[]``).
    :type exclude: ``list``
    :param bracket_wrap: wrap in brackets.
    :type bracket_wrap: ``bool``
    :return: dataframe with columns lists converted to strings.
    :rtype: ``Pandas DataFrame``
    """
    allowed = [['list'], ['list', 'nan']] if allow_nan else [['list']]

    # Find allowed columns
    cols_to_prettify = \
        [c for c in data_frame.columns if sorted(type_in_series(data_frame[c])) in allowed and c not in exclude]

    if len(cols_to_prettify) == 0:
        return data_frame
    else:
        return pandas_list_column_to_str(data_frame, cols_to_prettify, join_on, bracket_wrap)


def items_null(element):
    """

    Check if an object is a NaN, including all the elements in an iterable.

    :param element: a python object.
    :type element: ``any``
    :return: assessment of whether or not `element` is a NaN.
    :rtype: ``bool``
    """
    if isinstance(element, (list, tuple, type(np.array))):
        return True if all(pd.isnull(i) for i in element) else False
    else:
        return pd.isnull(element)


def pandas_null_drop(data_frame, subset=None):
    """

    Drop rows with NaNs of all, or a subset of, the dataframe's columns.
    (Can handle iterables which only contain NaNs).

    :param data_frame: a dataframe.
    :type data_frame: ``Pandas DataFrame``
    :param subset: a subset of columns. Defaults to None, which will apply to all columns.
    :type subset: ``iterable``
    :return: dataframe with NaN dropped.
    :rtype: ``Pandas DataFrame``
    """
    # Fill Empty cells with nans
    data_frame = data_frame.fillna(np.NaN)

    if subset is None:
        data_frame = data_frame.dropna()
    else:
        for c in subset:
            data_frame = data_frame[~data_frame[c].map(items_null)]

    return data_frame.reset_index(drop=True)

# ----------------------------------------------------------------------------------------------------------
# Printing Suit
# ----------------------------------------------------------------------------------------------------------

def _padding(s, amount, justify):
    """

    Add padding to a string.

    :param s: a string.
    :type s: ``str``
    :param amount: the amount of white space to add.
    :type amount: ``float`` or ``int``
    :param justify: the justification of the resultant text. Must be one of: 'left' or 'center'.
    :type justify: ``str``
    :return: `s` justified, or as passed if `justify` is not one of: 'left' or 'center'.
    :rtype: ``str``
    """
    pad = ' ' * amount
    if justify == 'left':
        return "%s%s" % (pstr(s), pad)
    elif justify == 'center':
        return "%s%s%s" % (pad[:int(amount/2)], pstr(s), pad[int(amount/2):])
    else:
        return s

def _pandas_series_alignment(pandas_series, justify):
    """

    Align all items in a pandas series.

    :param pandas_series: a pandas series.
    :type pandas_series: ``Pandas Series``
    :param justify: the justification of the resultant text. Must be one of: 'left', 'right' or 'center'.
    :type justify: ``str``
    :return: aligned series
    :rtype: ``str``
    """
    if justify == 'right':
        return pandas_series
    longest_string = max([len(s) for s in pandas_series.astype('unicode')])
    return [_padding(s, longest_string - len(s), justify) if not items_null(s) else s for s in pandas_series]


def align_pandas(data_frame, to_align='right'):
    """

    Align the columns of a Pandas DataFrame by adding whitespace.

    :param data_frame: a dataframe.
    :type data_frame: ``Pandas DataFrame``
    :param to_align: 'left', 'right', 'center' or a dictionary of the form: ``{'Column': 'Alignment'}``.
    :type to_align: ``str``
    :return: dataframe with aligned columns.
    :rtype: ``Pandas DataFrame``
    """
    if isinstance(to_align, dict):
        alignment_dict = to_align
    elif to_align.lower() in ['left', 'right', 'center']:
        alignment_dict = dict.fromkeys(data_frame.columns, to_align.lower())
    else:
        raise ValueError("to_align must be either 'left', 'right', 'center', or a dict.")

    for col, justification in alignment_dict.items():
        data_frame[col] = _pandas_series_alignment(data_frame[col], justification)

    return data_frame


def pandas_print_full(pd_df, full_rows = True, full_cols = True):
    """

    Print *all* of a Pandas DataFrame.

    :param pd_df: DataFrame to printed in its entirety.
    :type pd_df: ``Pandas DataFrame``
    :param full_rows: print all rows if True. Defaults to True.
    :type full_rows: ``bool``
    :param full_cols: print all columns side-by-side if True. Defaults to True.
    :type full_cols: ``bool``
    """
    if full_rows: pd.set_option('display.max_rows', len(pd_df))
    if full_cols: pd.set_option('expand_frame_repr', False)

    # Print the data frame
    print(pd_df)

    if full_rows: pd.reset_option('display.max_rows')
    if full_cols: pd.set_option('expand_frame_repr', True)


def pandas_pretty_print(data_frame, col_align='right', header_align='center', full_rows=True, full_cols=True):
    """

    Pretty Print a Pandas DataFrame.

    :param data_frame: a dataframe.
    :type data_frame: ``Pandas DataFrame``
    :param col_align: 'left', 'right', 'center'' or a dictionary of the form: ``{'Column': 'Alignment'}``.
    :type col_align: ``str`` or ``dict``
    :param header_align: alignment of headers. Must be one of: 'left', 'right', 'center'.
    :type header_align: ``str`` or ``dict``
    :param full_rows: print all rows.
    :type full_rows: ``bool``
    :param full_cols: print all columns.
    :type full_cols: ``bool``
    """
    aligned_df = align_pandas(data_frame, col_align)
    pd.set_option('colheader_justify', header_align)
    pandas_print_full(aligned_df.fillna(""), full_rows, full_cols)
    pd.set_option('colheader_justify', 'right')



















